/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  Copyright (c) Jack Botner and licensed under the MIT license; see
  http://opensource.org/licenses/mit-license.php

  nokia timer by Jack Botner

  interrupt.c

  This module contains all the interrupt handlers, including
  button debouncing logic.
  
  20201102 Fix sluggish rotary encoder response  
  20211215 Move encoder processing to main
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"

// Global variables

extern volatile uint8_t  ucEventFlags;		// system event flags

/*-----------------------------------------------------------------
  ISR(TIMER1_OVF_vect)

  Timer/Counter 1 overflow interrupt handler
  Used to provide 1/3 second intervals for reading and displaying
  the RTC. Also provides 1 second events for analyzing the timer
  state.

  System clock frequency = 8 mHz
  System clock period = 125 ns
  Prescaler = 256
  Timer count period = 32 us
  Counts = 10417
  Interrupt period = 1/3 sec

  The idea is to refresh the display frequently enough so that
  the seconds appear to increment regularly.

  20210103 Increase analyze state to 2/3 sec from 1 sec
-----------------------------------------------------------------*/
ISR(TIMER1_OVF_vect)
{
  static uint8_t ucCtr=0;

  set_bit( ucEventFlags, EVENT_READ_RTC );

  ++ucCtr;
  if ( ucCtr == 2 )
  {
	// Initiate a system state analysis every 2/3 second
	set_bit( ucEventFlags, EVENT_ANALYZE_STATE );
    ucCtr = 0;
  }

  TCNT1 = INIT_TIMER1;
}

/*-----------------------------------------------------------------
  ISR(TIMER2_OVF_vect)

  Timer 2 overflow interrupt handler - tests for button presses

  System clock frequency = 8 mHz
  System clock period = 125 ns
  Prescaler = 128
  Timer count period = 16 us
  Counts = 250
  Interrupt period = 4 ms
  
  20201226 Fix sluggish rotary encoder response ...
  20210103 Increase timer period from 1 to 4 ms to simplify code
------------------------------------------------------------------*/
ISR(TIMER2_OVF_vect)
{
  set_bit( ucEventFlags, EVENT_DEBOUNCE_IO );    // for encoder I/O
  TCNT2 = INIT_TIMER2;
}
