/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  Copyright (c) Jack Botner and licensed under the MIT license; see
  http://opensource.org/licenses/mit-license.php

  nokia timer by Jack Botner

  main.c

  This program implements an appliance timer with the following hardware:
  - Nokia 5110 LCD display 48 x 84 pixels (6 lines, 14 chars small font)
  - rotary encoder with pushbutton,
    EC11-1B-18T (Changzhou Xinze Electronic Co.) or
	  CTS series 290 p/n 290VAB0R201B2
  - a 5VDC relay that controls the external device
  - NJU6355 real time clock
  - 2 LEDs to indicate timer1/2 on or off.

  Timer1 uses a table of on/off times which change every week to
  account for seasonal changes. The idea is to turn the device on
  at dusk and off at dawn. If you live at a different latitude or
  time zone, the table must be customized. Resources are available on
  the internet to develop a customized table for your location. See
  function set_current_programmed_on_off_times in main.c for more
  information.

  Timer2 uses a programmable on/off time, plus can apply to a single
  day of the week or all days. When programming, 0 means all days of
  the week. 1 = Sunday, 2 = Monday, etc.

  Note that function isdaylighttime in calendar.c is coded for eastern
  standard/daylight time (North America). You may have to change that
  code too.

  The clock automatically corrects the time when standard time
  changes to daylight time, and vice versa.

  Internally time is kept in 24 hour format. Times are also programmed
  in 24 hour format. However it is displayed in 12 hour format; you
  can change that if you like.

  Processor: ATMega88PA
      Clock: Internal, 8.0 MHz

  20161208 Adapted from appliance timer project
  20201102 Fix sluggish rotary encoder response
  20211215 Move encoder processing to main
  20210122 Don't display seconds
-----------------------------------------------------------------*/
#include <stdint.h>
#include <avr/io.h>
//#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#define F_CPU 8000000UL
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "calendar.h"
#include "n3310_lcd.h"
#include "nju6355.h"

// Global Variables

RTCPARMS       stRtc;
MY_EEPROM_DATA stEepromVars;		    // variables kept in EEPROM

volatile uint8_t ucEventFlags=0;		// program event flags
volatile uint8_t wasDST;				    // previous DST state

volatile uint8_t ucTempOverrideOn=0;		// user wants the device on (or off)

// Timer1 parameters (from table)
static uint8_t	ucPrepgmHoursOn, ucPrepgmHoursOff;
static uint8_t	ucPrepgmMinsOn,  ucPrepgmMinsOff;

// Internal Functions

static void check_DST( void );
static uint8_t analyze_state( void );
static void format_and_display_time(void);
static void set_current_programmed_on_off_times(void);
static void system_init(void);
static uint8_t testTimer( uint16_t uiClock, uint16_t uiTimeOn, 
                          uint16_t uiTimeOff, uint8_t ucDayofweek );

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  uint8_t ii, ucOn;

  system_init();          		// initialize system

  set_bit( PORTD, PORTD6 );		// power up the LCD
  for ( ii=0; ii<2; ++ii )		// 50 ms
	_delay_loop_2( 50000 );		// 25 ms
  LCD_init();					      // initialize the LCD
  LCD_clear();
  LCD_write_string( 1, 1, "Timer 1.0", MENU_NORMAL );
   
  rtc_init();					    // initialize the rtc
  ucEventFlags = 0;				// clear all events

  // Get the saved state variables from eeprom
  eeprom_read_block( &stEepromVars, (void *) EEPROM_BASE, sizeof( MY_EEPROM_DATA ) );
  if ( stEepromVars.ucOnTimeHH > 23 ||		// probably EEPROM not initialized
	   stEepromVars.ucOnTimeMM > 59 ||
	   stEepromVars.ucOffTimeHH > 23 ||
	   stEepromVars.ucOffTimeMM > 59 )
  {
    stEepromVars.ucOnTimeHH = stEepromVars.ucOnTimeMM = 0;
    stEepromVars.ucOffTimeHH = stEepromVars.ucOffTimeMM = 0;
    stEepromVars.ucDayofWeek = 0;
	  stEepromVars.ucTimer1Enabled = stEepromVars.ucTimer2Enabled = 0;
	  set_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
  }
  
  // Initialize the preprogrammed on/off times
  set_current_programmed_on_off_times();

  // Initialize the DST flag
  wasDST = isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, 
                           stRtc.rtc_day,  stRtc.rtc_hours );

  for ( ii=0; ii<80; ++ii )		// 2 s
	_delay_loop_2( 50000 );		// 25 ms

  LCD_clear();
 
 if ( stEepromVars.ucTimer1Enabled  )
   set_bit( TIMER_LED_PORT, TIMER1_PORT_OUT );
 else
   clear_bit( TIMER_LED_PORT, TIMER1_PORT_OUT );
  
  if ( stEepromVars.ucTimer2Enabled )
    set_bit( TIMER_LED_PORT, TIMER2_PORT_OUT );
  else
    clear_bit( TIMER_LED_PORT, TIMER2_PORT_OUT );

  while(1)
  {
    if ( test_bit( ucEventFlags, EVENT_DEBOUNCE_IO ) )
    {
      encoder();      // check for encoder I/O
      clear_bit( ucEventFlags, EVENT_DEBOUNCE_IO );
    }

    if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
	{
	  main_menu();
	  clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
	}

	if ( test_bit( ucEventFlags, EVENT_ENCODER_UP ) ||
	     test_bit( ucEventFlags, EVENT_ENCODER_DOWN ) )
	{
	  // In run mode the encoder rotation is used to 
	  // turn the device on or off.
	  ucTempOverrideOn = ( ucTempOverrideOn ) ? 0 : 1;

	  clear_bit( ucEventFlags, EVENT_ENCODER_UP );
	  clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
	}

  	if ( test_bit( ucEventFlags, EVENT_READ_RTC ) )
	{
	  rtc_read();
	  format_and_display_time();
	  clear_bit( ucEventFlags, EVENT_READ_RTC );
	}

    if ( test_bit( ucEventFlags, EVENT_UPDATE_EEPROM ) )
	{
      eeprom_update_block( &stEepromVars, (void *) EEPROM_BASE, sizeof( MY_EEPROM_DATA ) );
      clear_bit( ucEventFlags, EVENT_UPDATE_EEPROM );
    }

    if ( test_bit( ucEventFlags, EVENT_ANALYZE_STATE ) )
    {
      ucOn = analyze_state();
      if ( ucOn )
      {
        // If the relay is off, turn it on
		if ( !test_bit( RELAY_PORT_IN, RELAY_PIN ) )
		  set_bit( RELAY_PORT_OUT, RELAY_PIN );
      }
      else
      {
        // If the relay is on, turn it off
        if ( test_bit( RELAY_PORT_IN, RELAY_PIN ) )
          clear_bit( RELAY_PORT_OUT, RELAY_PIN );
      }

      check_DST();      // corrects the clock when DST changes
      clear_bit( ucEventFlags, EVENT_ANALYZE_STATE );
      }
  }
}

/*-----------------------------------------------------------------
  format_and_display_time()

  Display the time on the LCD: HH MM, large font
-----------------------------------------------------------------*/
static void format_and_display_time()
{
  static volatile unsigned char ucLastHr=0;
  char	        buffer[12], temp[4];
  unsigned char ucHr, ucX;

  ucHr = stRtc.rtc_hours;
  // Since the clock is in 24 hour format, make adjustment
  if ( !ucHr )
    ucHr = 12;
  else
    if ( ucHr > 12 )
      ucHr = ucHr - 12;

  // Shift right 1 position when there is only one hour digit
  // to show.
  ucX = ( ucHr < 10 ) ? 2 : 1;

  utoa( (uint16_t) ucHr, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	  buffer[0] = '0';
	  buffer[1] = temp[0];
  }
  else
  {
	  buffer[0] = temp[0];
	  buffer[1] = temp[1];
  }
  buffer[2] = '.';

  utoa( (uint16_t) stRtc.rtc_mins, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	  buffer[3] = '0';
	  buffer[4] = temp[0];
  }
  else
  {
	  buffer[3] = temp[0];
	  buffer[4] = temp[1];
  }
  buffer[5] = '\0';
#if 0
  buffer[5] = '.';

  utoa( (uint16_t) stRtc.rtc_secs, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	  buffer[6] = '0';
	  buffer[7] = temp[0];
  }
  else
  {
	  buffer[6] = temp[0];
	  buffer[7] = temp[1];
  }

  buffer[8] = '\0';
#endif

  if ( ucLastHr != ucHr )
  {
	  // change of hour can cause shifting
	  LCD_clear();
	  ucLastHr = ucHr;
  }
  
  ucX += 12;
  LCD_write_string_big( ucX, 1, buffer, MENU_NORMAL );
}

/*-----------------------------------------------------------------
  analyze_state()

  See if the device should be on or off. Called once a second.
  Timer1 uses table driven logic. Timer2 uses user-programmed
  day/time.

  Returns: 1 (on) or 0 (off)
-----------------------------------------------------------------*/
static uint8_t analyze_state()
{
  uint8_t   ucTimer1State=0, ucTimer2State=0;
  uint16_t	uiTimeOn, uiTimeOff, uiClock;
  uint8_t   ucDayofWeek;

  uiClock = ( (uint16_t) stRtc.rtc_hours * 60 ) + (uint16_t) stRtc.rtc_mins;

  // Process timer1

  if ( stRtc.rtc_secs == 30 )
  {
    // Arbitrarily at 30 seconds each minute refresh the
    // preprogram settings.
    set_current_programmed_on_off_times();
  }

  if ( stEepromVars.ucTimer1Enabled )
  {
    // Use the on and off times from our table
    uiTimeOn  = ( (uint16_t) ucPrepgmHoursOn * 60 )  + (uint16_t) ucPrepgmMinsOn;
    uiTimeOff = ( (uint16_t) ucPrepgmHoursOff * 60 ) + (uint16_t) ucPrepgmMinsOff;
    ucDayofWeek = 0;

    ucTimer1State = testTimer( uiClock, uiTimeOn, uiTimeOff, ucDayofWeek );

    // If the on or off time is now and the temporary override is on, cancel it.
    if ( uiClock == uiTimeOn || uiClock == uiTimeOff )
      ucTempOverrideOn = 0;
  }
        
  // Process timer2

  if ( stEepromVars.ucTimer2Enabled )
  {
	  uiTimeOn  = ( (uint16_t) stEepromVars.ucOnTimeHH * 60 )  + (uint16_t) stEepromVars.ucOnTimeMM;
	  uiTimeOff = ( (uint16_t) stEepromVars.ucOffTimeHH * 60 ) + (uint16_t) stEepromVars.ucOffTimeMM;
    ucDayofWeek = stEepromVars.ucDayofWeek;

    ucTimer2State = testTimer( uiClock, uiTimeOn, uiTimeOff, ucDayofWeek );

    // If the on or off time is now and the temporary override is on, cancel it.
    if ( uiClock == uiTimeOn || uiClock == uiTimeOff )
      ucTempOverrideOn = 0;
  }

  // Process the override setting

  if ( !ucTempOverrideOn )
  {
    // The override is off so return 1 if either timer is on
    if ( ucTimer1State || ucTimer2State )
      return 1;
    else
      return 0;
  }

  // The override is on, so we want to return the
  // reverse logic from the timers

  if ( ucTimer1State || ucTimer2State )
    return 0;

  return 1;
}

/*-----------------------------------------------------------------
  testTimer()

  Uses the clock and the timer parameters to decide if the device
  should be on or off.

  Input: clock time
         timer on/off parameters

  Returns: 1 (on) or 0 (off)
-----------------------------------------------------------------*/
static uint8_t testTimer( uint16_t uiClock, uint16_t uiTimeOn, 
                          uint16_t uiTimeOff, uint8_t ucDayofweek )
{
  if ( uiTimeOn == uiTimeOff )
  // On/Off times have not been programmed yet (or improperly programmed)
  return 0;

  if ( ucDayofweek )          // specific day requested
    if ( ucDayofweek != stRtc.rtc_dayweek )
      return 0;

  // See whether the device should be on or off. Note that the
  // off time may be less than the on time, which means that it
  // is tomorrow time.
  if ( uiTimeOff > uiTimeOn )
  {
    // Off same day
    if ( uiClock >= uiTimeOn && uiClock < uiTimeOff )
      return 1;
  }
  else
  {
    // Off next day
    if ( uiClock >= uiTimeOn || uiClock < uiTimeOff )
      return 1;
  }

  return 0;
}

/*-----------------------------------------------------------------
  check_DST()

  Called once a second. We need to set the time ahead one hour
  in the spring and one hour back in the fall. Special logic is
  required in the fall to keep us from repeating the hour change.
-----------------------------------------------------------------*/
static void check_DST()
{
  static volatile uint8_t holdoff=0;
  uint8_t    isDST;

  // Do only when seconds and minutes = 0
  if ( stRtc.rtc_secs || stRtc.rtc_mins ) return;

  if ( holdoff )
  {
    // This flag tells us the time was set back one hour recently.
	  // Wait until 3AM to prevent us from repeating, then clear the flag.
    if ( stRtc.rtc_hours == 3 )
	  holdoff = 0;

    return;
  }

  isDST = isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, 
                          stRtc.rtc_day,  stRtc.rtc_hours );

  if ( isDST == wasDST ) return;   // no change

  if ( isDST )
  {
    // We have just gone from standard time to DST.
	  // We need to put the clock ahead 1 hour.
	  ++stRtc.rtc_hours;
      rtc_write();
	  wasDST = isDST;
  }
  else
  {
    // We have just gone from DST to standard time.
	  // We need to put the clock back 1 hour.
	  --stRtc.rtc_hours;
    rtc_write();
	  wasDST = isDST;
    holdoff = 1;     // to keep us from doing this again in an hour
  }
}

/*-----------------------------------------------------------------
  set_current_programmed_on_off_times()

  Sets the current programmed on and off times from our tables and
  the current clock.

  Output: ucPrepgmHoursOn, ucPrepgmHoursOff, ucPrepgmMinsOn,
		  ucPrepgmMinsOff
-----------------------------------------------------------------*/
static void set_current_programmed_on_off_times()
{
  uint8_t   ucIndex;

  // Table for setting on and off times depending on time of year.
  // To keep it simple there is one entry for each week. All times
  // must be in "standard" time (not DST). These tables are for
  // Toronto, ON (43 deg. 46 min. N, 79 deg. 23 min. W, courtesy
  // http://aa.usno.navy.mil/data.

  static uint8_t ucOntimeHH[52] = 
  {
    16, 17, 17, 17, 17, 17, 17, 17, 18, 18, 18, 18, 18, 
	  18, 18, 19, 19, 19, 19, 19, 19, 19, 19, 20, 20, 20,
	  20, 20, 19, 19, 19, 19, 19, 19, 19, 18, 18, 18, 18,
	  18, 17, 17, 17, 17, 17, 16, 16, 16, 16, 16, 16, 16
  };
  static uint8_t ucOntimeMM[52] = 
  {
    50,  0,  5, 15, 25, 35, 45, 55,  5, 10, 20, 30, 40, 
	  45, 55,  5, 10, 20, 30, 35, 45, 50, 55,  0,  5,  5,
	   5,  0, 55, 50, 45, 35, 25, 15,  0, 50, 40, 25, 10,
	   0, 45, 35, 25, 15,  5, 55, 50, 45, 40, 40, 40, 45
  };
  static uint8_t ucOfftimeHH[52] = 
  {
     7,  7,  7,  7,  7,  7,  7,  7,  7,  6,  6,  6,  6, 
	   5,  5,  5,  5,  5,  5,  4,  4,  4,  4,  4,  4,  4,
	   4,  4,  4,  5,  5,  5,  5,  5,  5,  5,  6,  6,  6,
	   6,  6,  6,  6,  7,  7,  7,  7,  7,  7,  7,  7,  7
  };
  static uint8_t ucOfftimeMM[52] = 
  {
    50, 50, 45, 45, 35, 30, 20, 10,  0, 45, 35, 20, 10, 
	  55, 45, 30, 20, 10,  0, 50, 45, 40, 35, 30, 35, 40,
	  45, 50, 55,  0, 10, 20, 25, 35, 45, 50,  0,  5, 15,
	  25, 30, 40, 50,  0,  5,  5, 15, 25, 35, 40, 45, 50
  };

  // Use the on and off times from our table
  ucIndex = week_of_year( stRtc.rtc_year, stRtc.rtc_month, stRtc.rtc_day );
  ucPrepgmHoursOn  = ucOntimeHH[ucIndex];
  ucPrepgmHoursOff = ucOfftimeHH[ucIndex];
  ucPrepgmMinsOn   = ucOntimeMM[ucIndex];
  ucPrepgmMinsOff  = ucOfftimeMM[ucIndex];

  if ( isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, stRtc.rtc_day,
						stRtc.rtc_hours ) )
  {
    // Adjust the table values for DST
	++ucPrepgmHoursOn;
	++ucPrepgmHoursOff;
	// It's possible to overflow if hours are 23, but there's no 23 hours
	// in my table. If it's a possibility then test for overflow.
  }
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (output) LCD D/C
         PB1 (output) LCD reset
         PB2 (output) LCD chip enable
         PB3 (output) ISP MOSI/LCD data in
         PB4 (input)  ISP MISO
         PB5 (output) ISP SCK/LCD clock
		 PB6 (output) not used
         PB7 (output) not used

  PORTC: PC0 (input) Encoder "A"
		 PC1 (input) Encoder "B"
		 PC2 (input) Encoder pushbutton
		 PC3 (input) not used
         PC4 (output) Timer 1 LED
         PC5 (output) Timer 2 LED
         PC6 (output) ISP !Reset 

  PORTD: PD0 (output) RTC data
	     PD1 (output) RTC I/O
		 PD2 (output) RTC CLK
		 PD3 (output) RTC CE
		 PD4 (input) not used
		 PD5 (output) Relay Control
         PD6 (input) not used
         PD7 (input) not used
-----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set prescaler to 1

  // Set up the I/O ports directions, pull-ups
  DDRB  = 0b00101111;		// port B directions
  PORTB = 0b11010000;       // port B initialization, pullups

  DDRC  = 0b01110000;		// pc0..6
  PORTC = 0b00111111;       // initialize PortC

  DDRD  = 0b00101111;		// pd0..3=output, pd4=input, pd5..7=output
  PORTD = 0b11000000;       // initialize PortD

  // Initialize Timer1.
  // Timer1 is used to create 1 s intervals for scheduling events
  TCCR1A = 0b00000000;		// normal operation, no OC1A, no OC1B
  TCCR1B = 0b00000100;      // internal clock, prescaler = 256
  TCNT1 = INIT_TIMER1;
  set_bit( TIMSK1, TOIE1 );	// enable timer/counter1 overflow interrupt

  // Initialize Timer2. Timer2 is used to generate interrupts
  // every 1 millisecond for switch debouncing.
  TCCR2A = 0b00000000;		// normal port operation
  TCCR2B = 0b00000101;		// normal operation, prescaler=128
  TCNT2 = INIT_TIMER2;
  set_bit( TIMSK2, TOIE2 );	// enable timer/counter2 overflow interrupt

  sei();					// enable global interrupts
}

#ifdef DIAGNOSTICS
/*-----------------------------------------------------------------
  diagnostics()
-----------------------------------------------------------------*/
void diagnostics()
{
  char cWork[16], cWk[8];

  strcpy( cWork, "Timer1 ");
  if ( stEepromVars.ucTimer1Enabled )
    strcat( cWork, "ON" );
  else
    strcat( cWork, "OFF" );
  LCD_write_string( 0, 0, cWork, MENU_NORMAL );

  strcpy( cWork, "ON = ");
  utoa( (uint16_t) ucPrepgmHoursOn, cWk, 10 );
  strcat( cWork, cWk );
  strcat( cWork, ", " );
  utoa( (uint16_t) ucPrepgmMinsOn, cWk, 10 );
  strcat( cWork, cWk );
  LCD_write_string( 0, 1, cWork, MENU_NORMAL );

  strcpy( cWork, "OFF = ");
  utoa( (uint16_t) ucPrepgmHoursOff, cWk, 10 );
  strcat( cWork, cWk );
  strcat( cWork, ", " );
  utoa( (uint16_t) ucPrepgmMinsOff, cWk, 10 );
  strcat( cWork, cWk );
  LCD_write_string( 0, 2, cWork, MENU_NORMAL );

  while ( 1 )
  {
	  _delay_loop_2( 50000 );		// 25 ms
    if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
      break;
  }
  clear_bit( ucEventFlags, EVENT_ENCODER_UP );
  clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
	clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
	LCD_clear();
}
#endif
